// EdgeMask - builds the edge map of a video plane based on selectable
// differential operator
// EdgeMask (C) 2003 Kurosu (kurosu@inforezo.org)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .


#include <string.h>
#include <stdio.h>
#include "GenericTools.h"
#include "DEdgeMask.h"

int DEdgeMask::ConvertCharToInt(const char c)
{
	switch(c)
	{
		case '1' : return 1;
		case '2' : return 2;
		case '3' : return 3;
		case '4' : return 4;
		case '5' : return 5;
		case '6' : return 6;
		case '7' : return 7;
		case '8' : return 8;
		case '9' : return 9;
		default : return 0;
	}
}

int DEdgeMask::ConvertStringToInt(const char *string, int end)
{
	int answer = 0;
	int decade = 1;
	for ( int pointer = end-1; pointer >= 0; pointer--)
	{
		if (( string[pointer] == '-') && (pointer == 0))
			answer = -answer;
		else 
			answer += ConvertCharToInt(string[pointer]) * decade;
		decade *= 10;
	}
	return answer;
}

bool DEdgeMask::AnalyseString(const char *string, int *vect, IScriptEnvironment* env)
{
	int length = strlen(string);
	char separator[12] = {'0','1','2','3','4','5','6','7','8','9','-',0};

    int pos;
	int count = 0;
	int prevpos = 0;
	bool answer = false;
	while ( (pos = strspn(string + prevpos,separator)) > 0)
	{
		prevpos = prevpos + pos + strcspn(string+prevpos+pos,separator);
		count++;
	}
	
	if (count != 9) env->ThrowError("DEdgeMask: matrix must have 9 values");
	
	prevpos = 0;
	count = 0;
	while ( (pos = strspn(string+prevpos,separator)) > 0)
	{
		vect[count] = ConvertStringToInt(string + prevpos, pos);
		answer |= ( vect[count] < 0 );
		prevpos = prevpos + pos + strcspn(string+prevpos+pos,separator);
		count++;
	}

	vect[count] = ConvertStringToInt(string + prevpos, length - prevpos);
	answer |= ( vect[count] < 0 );

	return answer;
}

DEdgeMask::DEdgeMask(PClip _child, unsigned int thY1, unsigned int thY2,
				   unsigned int thC1, unsigned int thC2, const char *s,
				   int y, int u, int v, bool useMMX, bool setDivisor, int div, IScriptEnvironment* env) :
GenericVideoFilter(_child), Yth1(thY1), Yth2(thY2), Cth1(thC1), Cth2(thC2), Y(y), V(v), U(u)
{
	if (!vi.IsYV12())
		env->ThrowError("DEdgeMask: Needs YV12 data");
	
	if (thY1>255)
		env->ThrowError("DEdgeMask: Threshold Y1 not in the range [0..255]");
	if (thY2>255)
		env->ThrowError("DEdgeMask: Threshold Y2 not in the range [0..255]");
	if (thC1>255)
		env->ThrowError("DEdgeMask: Threshold C1 not in the range [0..255]");
	if (thC2>255)
		env->ThrowError("DEdgeMask: Threshold C2 not in the range [0..255]");

	AnalyseString(s,array,env);

	if ( setDivisor ) divisor = div;
	else {
		divisor = 0;
		for ( int i = 0; i < 9; i++ )
			divisor += (array[i] > 0) ? array[i] : 0;
	}

	if (!Verify(Y)) env->ThrowError("DEdgeMask: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("DEdgeMask: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("DEdgeMask: Unvalid operation mode on V");


}

void DEdgeMask::BuildMask (const unsigned char *srcp, unsigned char *dstp, const int src_pitch,
							const int dst_pitch, int row_size, const int height,
							const unsigned char th1, const unsigned char th2)
{
	const unsigned char *su = srcp + 1;
	const unsigned char *s = srcp + src_pitch + 1;
	const unsigned char *sd = srcp + 2 * src_pitch + 1;
	unsigned char *d = dstp + dst_pitch;
	int srcmod = src_pitch + 2 - row_size;
	int sum;

	for ( int j = 1; j < height - 1; j++ )
	{
		for ( int i = 1; i < row_size - 1; i++ )
		{
			sum = MyAbs(array[0] * su[-1] + array[1] * su[0] + array[2] * su[1] +
						array[3] * s[-1]  + array[4] * s[0]  + array[5] * s[1]  +
						array[6] * sd[-1] + array[7] * sd[0] + array[8] * sd[1]) / divisor;

			d[i] = ( sum <= th1 ) ? 0 : ( sum > th2 ) ? 255 : sum;
			s++; su++; sd++;
		}
		d[0] = d[1];
		d[row_size-1] = d[row_size-2];
		d += dst_pitch;
		s += srcmod;
		su += srcmod;
		sd += srcmod;
	}
	
	for ( int x = 0; x < row_size; x++ )
	{
		dstp[x] = dstp[x+dst_pitch];
		d[x] = d[x-dst_pitch];
	}

}

PVideoFrame __stdcall DEdgeMask::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			src = child->GetFrame(n, env);
	env->MakeWritable(&src);
	const unsigned char *srcp_y = src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcp_u = src->GetReadPtr(PLANAR_U);
	const unsigned char *srcp_v = src->GetReadPtr(PLANAR_V);
	const int			src_pitch_y = src->GetPitch(PLANAR_Y);
	const int			src_pitch_uv = src->GetPitch(PLANAR_U);

	PVideoFrame			src_prev = child->GetFrame(n, env);
	const unsigned char *src_prev_y = src_prev->GetReadPtr(PLANAR_Y);
	const unsigned char *src_prev_u = src_prev->GetReadPtr(PLANAR_U);
	const unsigned char *src_prev_v = src_prev->GetReadPtr(PLANAR_V);
	const int			src_prev_pitch_y = src_prev->GetPitch(PLANAR_Y);
	const int			src_prev_pitch_uv = src_prev->GetPitch(PLANAR_U);
	
	PVideoFrame			der = env->NewVideoFrame(vi);
	unsigned char		*derp_y = der->GetWritePtr(PLANAR_Y);
	unsigned char		*derp_u = der->GetWritePtr(PLANAR_U);
	unsigned char		*derp_v = der->GetWritePtr(PLANAR_V);
	const int			der_pitch_y = der->GetPitch(PLANAR_Y);
	const int			der_pitch_uv = der->GetPitch(PLANAR_U);
	
	const int			row_size = src->GetRowSize(PLANAR_Y);
	const int			height = src->GetHeight(PLANAR_Y);
	
	switch (Y)
	{
	case OPERATION:
		BuildMask(srcp_y, derp_y, src_pitch_y, der_pitch_y, row_size,
					height, Yth1, Yth2);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_y,der_pitch_y,srcp_y,src_pitch_y,row_size,height);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_y, -Y, row_size, height, 0, 0, der_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		BuildMask(srcp_u, derp_u, src_pitch_uv, der_pitch_uv, row_size>>1,
					height>>1, Cth1, Cth2);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_u, der_pitch_uv, srcp_u, src_pitch_uv,
					(row_size>>1), (height>>1));
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_u, -U, row_size / 2, height / 2, 0, 0, der_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		BuildMask(srcp_v, derp_v, src_pitch_uv, der_pitch_uv, row_size>>1,
					height>>1, Cth1, Cth2);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_v, der_pitch_uv, srcp_v, src_pitch_uv,
					(row_size>>1),(height>>1));
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_v, -V, row_size / 2, height / 2, 0, 0, der_pitch_uv);
		break;
	}
	
	return der;
}
