// EdgeMask - builds the edge map of a video plane based on selectable
// differential operator
// EdgeMask (C) 2003 Kurosu (kurosu@inforezo.org)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .


#include <string.h>

#include "EdgeMask.h"
#include "GenericTools.h"
#include "EM_func.h"

EdgeMask::EdgeMask(PClip _child, unsigned int thY1, unsigned int thY2,
				   unsigned int thC1, unsigned int thC2, const char *string,
				   int y, int u, int v, bool useMMX, IScriptEnvironment* env) :
GenericVideoFilter(_child), Yth1(thY1), Yth2(thY2), Cth1(thC1), Cth2(thC2), Y(y), V(v), U(u)
{
	if (!vi.IsYV12())
		env->ThrowError("EdgeMask: Needs YV12 data");
	
	if (thY1>255)
		env->ThrowError("EdgeMask: Threshold Y1 not in the range [0..255]");
	if (thY2>255)
		env->ThrowError("EdgeMask: Threshold Y2 not in the range [0..255]");
	if (thC1>255)
		env->ThrowError("EdgeMask: Threshold C1 not in the range [0..255]");
	if (thC2>255)
		env->ThrowError("EdgeMask: Threshold C2 not in the range [0..255]");

	/* Check mode and if width is MOD8 */
	if (!lstrcmpi(string,"line"))
	{
		if ((env->GetCPUFlags() & CPUF_MMX) && !(vi.width%16) && useMMX)
			BuildMask = Line_C;
		else BuildMask = Line_C;
	}
	else if (!lstrcmpi(string,"roberts"))
	{
		if ((env->GetCPUFlags() & CPUF_MMX) && !(vi.width%16) && useMMX)
			BuildMask = Roberts_MMX;
		else BuildMask = Roberts_C;
	}
	else if (!lstrcmpi(string,"sobel"))
	{
		if ((env->GetCPUFlags() & CPUF_MMX) && !(vi.width%16) && useMMX)
			BuildMask = Sobel_MMX;
		else BuildMask = Sobel_C;
	}
	else if (!lstrcmpi(string,"special"))
	{
		if ((env->GetCPUFlags() & CPUF_MMX) && !(vi.width%16) && useMMX)
			BuildMask = Special_MMX;
		else BuildMask = Special_C;
	}
	else if (!lstrcmpi(string,"cartoon"))
	{
		if ((env->GetCPUFlags() & CPUF_MMX) && !(vi.width%16) && useMMX)
			BuildMask = Cartoon_MMX;
		else BuildMask = Cartoon_C;
	}
	else if (!lstrcmpi(string,"laplace"))
	{
		BuildMask = Laplace_C;
	}
	else env->ThrowError("EdgeMask: unvalid edge detection");
	
	if (!Verify(Y)) env->ThrowError("EdgeMask: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("EdgeMask: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("EdgeMask: Unvalid operation mode on V");
}

PVideoFrame __stdcall EdgeMask::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			src = child->GetFrame(n, env);

	const unsigned char *srcp_y = src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcp_u = src->GetReadPtr(PLANAR_U);
	const unsigned char *srcp_v = src->GetReadPtr(PLANAR_V);
	const int			src_pitch_y = src->GetPitch(PLANAR_Y);
	const int			src_pitch_uv = src->GetPitch(PLANAR_U);

	PVideoFrame			der = env->NewVideoFrame(vi);
	unsigned char		*derp_y = der->GetWritePtr(PLANAR_Y);
	unsigned char		*derp_u = der->GetWritePtr(PLANAR_U);
	unsigned char		*derp_v = der->GetWritePtr(PLANAR_V);
	const int			der_pitch_y = der->GetPitch(PLANAR_Y);
	const int			der_pitch_uv = der->GetPitch(PLANAR_U);
	
	const int			row_size = src->GetRowSize(PLANAR_Y);
	const int			height = src->GetHeight(PLANAR_Y);
	
	switch (Y)
	{
	case OPERATION:
		BuildMask(srcp_y, derp_y, src_pitch_y, der_pitch_y, row_size,
					height, Yth1, Yth2);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_y,der_pitch_y,srcp_y,src_pitch_y,row_size,height);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_y, -Y, row_size, height, 0, 0, der_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		BuildMask(srcp_u, derp_u, src_pitch_uv, der_pitch_uv, row_size>>1,
					height>>1, Cth1, Cth2);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_u, der_pitch_uv, srcp_u, src_pitch_uv,
					(row_size>>1), (height>>1));
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_u, -U, row_size / 2, height / 2, 0, 0, der_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		BuildMask(srcp_v, derp_v, src_pitch_uv, der_pitch_uv, row_size>>1,
					height>>1, Cth1, Cth2);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_v, der_pitch_uv, srcp_v, src_pitch_uv,
					(row_size>>1),(height>>1));
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_v, -V, row_size / 2, height / 2, 0, 0, der_pitch_uv);
		break;
	}
	
	return der;
}
