// Expand - set of tools for region growth of at best one pixel
// Expand (C) 2003 Kurosu (kurosu@inforezo.org)

// This work can also be referred to by the word AreaModifier

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "GenericTools.h"
#include "Expand.h"
#include "EXP_func.h"

Expand::Expand(PClip _child, int mode, int _X, int _Y, int _w, int _h,
			   int y, int u, int v, IScriptEnvironment* env) : 
GenericVideoFilter(_child), Y(y), V(v), U(u), w(_w), h(_h),
offX(_X & 0xFFFFFFFE), offY(_Y & 0xFFFFFFFE)
{
	if (!vi.IsYV12())
		env->ThrowError("Inflate: Needs YV12 data");
	
	if (!Verify(Y)) env->ThrowError("AreaModifier: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("AreaModifier: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("AreaModifier: Unvalid operation mode on V");

	if (offX<0 || offX>vi.width)
		env->ThrowError("AreaModifier: horizontal offset %i bigger than source width %i",
						offX, vi.width);
	if (offY<0 || offY>vi.height)
		env->ThrowError("AreaModifier: vertical offset %i bigger than source height %i",
						offY, vi.height);

	if (w < 0) w = vi.width;
	if (w > vi.width - offX) w = vi.width-offX;
	if (h < 0) h = vi.height;
	if (h > vi.height - offY) h = vi.height-offY;
	w &= 0xFFFFFFFE;
	h &= 0xFFFFFFFE;
	
	switch(mode)
	{
		case MODE_EXPAND:
			Process = Expand_C;
			break;
		case MODE_INPAND:
			Process = Inpand_C;
			break;
		case MODE_INFLATE:
			Process = Inflate_C;
			break;
		case MODE_DEFLATE:
			Process = Deflate_C;
			break;
		default:
			env->ThrowError("AreaModifier: Invalid operation mode");
			break;
	}
}

PVideoFrame __stdcall Expand::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			src = child->GetFrame(n, env);
	const unsigned char *srcp_y = src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcp_u = src->GetReadPtr(PLANAR_U);
	const unsigned char *srcp_v = src->GetReadPtr(PLANAR_V);
	const int			src_pitch_y = src->GetPitch(PLANAR_Y);
	const int			src_pitch_uv = src->GetPitch(PLANAR_U);
	
	PVideoFrame			der = env->NewVideoFrame(vi);
	unsigned char		*derp_y = der->GetWritePtr(PLANAR_Y);
	unsigned char		*derp_u = der->GetWritePtr(PLANAR_U);
	unsigned char		*derp_v = der->GetWritePtr(PLANAR_V);
	const int			der_pitch_y = der->GetPitch(PLANAR_Y);
	const int			der_pitch_uv = der->GetPitch(PLANAR_U);
	
	const int			row_size = src->GetRowSize(PLANAR_Y);
	const int			height = src->GetHeight(PLANAR_Y);
	
	switch (Y)
	{
	case OPERATION:
		Process(srcp_y + offX + offY*src_pitch_y, derp_y + offX + offY*der_pitch_y,
				src_pitch_y, der_pitch_y, w, h);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_y + offX + offY * der_pitch_y, der_pitch_y,
					srcp_y + offX + offY * src_pitch_y, src_pitch_y, w, h);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_y, -Y, w, h, offX, offY, der_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		Process(srcp_u + (offX>>1) + (offY>>1) * src_pitch_uv,
				derp_u + (offX>>1) + (offY>>1) * der_pitch_uv,
				src_pitch_uv, der_pitch_uv, w>>1, h>>1);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_u + (offX>>1) + (offY>>1) * der_pitch_uv, der_pitch_uv,
					srcp_u + (offX>>1) + (offY>>1) * src_pitch_uv, src_pitch_uv, w>>1, h>>1);
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_u, -U, w>>1, h>>1, offX>>1, offY>>1, der_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		Process(srcp_v + (offX>>1) + (offY>>1) * src_pitch_uv,
				derp_v + (offX>>1) + (offY>>1) * der_pitch_uv,
				src_pitch_uv, der_pitch_uv, w>>1, h>>1);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_v + (offX>>1) + (offY>>1) * der_pitch_uv, der_pitch_uv,
					srcp_v + (offX>>1) + (offY>>1) * src_pitch_uv, src_pitch_uv, w>>1, h>>1);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_v, -V, w>>1, h>>1, offX>>1, offY>>1, der_pitch_uv);
		break;
	}
	
	return der;
}