// YV12Subtract - expansion of AviSynth's Subtract
// YV12Subtract (C) 2003 Kurosu (kurosu@inforezo.org)
// Subtract (C) AviSynth's team

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "GenericTools.h"
#include "HysteresyMask.h"

HysteresyMask::HysteresyMask(PClip _child, PClip _child2, int y, int u, int v, IScriptEnvironment* env) :
GenericVideoFilter(_child), child2(_child2), Y(y), V(v), U(u)
{
	const VideoInfo& vi1 = _child->GetVideoInfo();
	const VideoInfo& vi2 = _child2->GetVideoInfo();
	if ( (vi1.width != vi2.width) || (vi1.height != vi2.height) )
		env->ThrowError("HysteresyMask: Image dimensions don't match");

	if (!vi1.IsYV12() || !vi2.IsYV12())
		env->ThrowError("HysteresyMask: requires YV12 colorspace");
	
	if (!Verify(Y)) env->ThrowError("YV12Substract: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("YV12Substract: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("YV12Substract: Unvalid operation mode on V");

	isInitialized = false;
}
HysteresyMask::~HysteresyMask()
{
	if (isInitialized)
	{
		delete[] Xarray;
		delete[] Yarray;
		delete[] used;
	}
}

void HysteresyMask::ExpandMask(const unsigned char *RMask, const unsigned char *EMask,
							unsigned char *ResultMask, const int row_size, const int height,
							const int RMPitch, const int EMPitch, const int ResMPitch, int x, int y)
{
	Xarray[0] = x;
	Yarray[0] = y;
	used[x + y * row_size] = 255;
	ResultMask[x + y * ResMPitch] = 255;
	int taille = 1;
	while ( taille > 0 )
	{
		taille--;
		int Xcurr = Xarray[taille];
		int Ycurr = Yarray[taille];
		for ( int a = -1; a < 2; a++ )
		{
			for ( int b = -1; b < 2; b++ )
			{
				if (( a + Xcurr >= 0 ) && ( a + Xcurr < row_size ) && ( b + Ycurr >= 0 )
					&& ( b + Ycurr < height ) && (used[Xcurr + a + (Ycurr + b) * row_size] == 0) &&
					( EMask[Xcurr + a + (Ycurr + b) * EMPitch] != 0 ))
				{
					Xarray[taille] = a + Xcurr;
					Yarray[taille] = b + Ycurr;
					used[Xcurr + a + (Ycurr + b) * row_size] = 255;
					ResultMask[Xcurr + a + (Ycurr + b) * ResMPitch] = 255;
					taille++;
				}
			}
		}
	}

}

void HysteresyMask::BuildMask(const unsigned char *RMask, const unsigned char *EMask,
							unsigned char *ResultMask, const int row_size, const int height,
							const int RMPitch, const int EMPitch, const int ResMPitch)
{
	memset(used,0,row_size*height);
	memset(ResultMask,0,ResMPitch*height);
	int pos = 0;
	for ( int j = 0; j < height; j++)
	{
		for ( int i = 0; i < row_size; i++ )
		{
			if (( used[i + j * row_size] == 0 ) && ( RMask[i + j * RMPitch] != 0 ) && ( EMask[i + j * EMPitch] != 0))
				ExpandMask(RMask, EMask, ResultMask, row_size, height, RMPitch, EMPitch,
							ResMPitch, i, j);
		}
	}
}


PVideoFrame __stdcall HysteresyMask::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			src1 = child->GetFrame(n, env);
	
	const unsigned char	*src1p_y = src1->GetWritePtr(PLANAR_Y);
	const unsigned char	*src1p_u = src1->GetWritePtr(PLANAR_U);
	const unsigned char	*src1p_v = src1->GetWritePtr(PLANAR_V);
	const int			src1_pitch_y = src1->GetPitch(PLANAR_Y);
	const int			src1_pitch_uv = src1->GetPitch(PLANAR_U);
	
	PVideoFrame			src2 = child2->GetFrame(n, env);
	const unsigned char	*src2p_y = src2->GetReadPtr(PLANAR_Y);
	const unsigned char	*src2p_u = src2->GetReadPtr(PLANAR_U);
	const unsigned char	*src2p_v = src2->GetReadPtr(PLANAR_V);
	const int			src2_pitch_y = src2->GetPitch(PLANAR_Y);
	const int			src2_pitch_uv = src2->GetPitch(PLANAR_U);

	PVideoFrame			der = env->NewVideoFrame(vi);
	unsigned char		*derp_y = der->GetWritePtr(PLANAR_Y);
	unsigned char		*derp_u = der->GetWritePtr(PLANAR_U);
	unsigned char		*derp_v = der->GetWritePtr(PLANAR_V);
	const int			der_pitch_y = der->GetPitch(PLANAR_Y);
	const int			der_pitch_uv = der->GetPitch(PLANAR_U);
	
	const int			row_size = src1->GetRowSize(PLANAR_Y);
	const int			height = src1->GetHeight(PLANAR_Y);

	if ( !isInitialized )
	{
		used = new unsigned char[row_size*height];
		Xarray = new int[row_size*height];
		Yarray = new int[row_size*height];
		isInitialized = true;
	}
	
	switch (Y)
	{
	case OPERATION:
		BuildMask(src1p_y, src2p_y, derp_y, row_size, height, src1_pitch_y, src2_pitch_y, der_pitch_y);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_y, der_pitch_y, src1p_y, src1_pitch_y, row_size, height);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_y, -Y, row_size, height, 0, 0, der_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		BuildMask(src1p_u, src2p_u, derp_u, row_size / 2, height / 2, src1_pitch_uv, src2_pitch_uv, der_pitch_uv);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_u, der_pitch_uv, src1p_u, src1_pitch_uv, row_size / 2, height / 2);
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_u, -U, row_size / 2, height / 2, 0, 0, der_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		BuildMask(src1p_v, src2p_v, derp_v, row_size / 2, height / 2, src1_pitch_uv, src2_pitch_uv, der_pitch_uv);
		break;
	case COPY_SOURCE:
		env->BitBlt(derp_v, der_pitch_uv, src1p_v, src1_pitch_uv, row_size / 2, height / 2);
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(derp_v, -V, row_size / 2, height / 2, 0, 0, der_pitch_uv);
		break;
	}
	
	return der;
}
