// YV12LUT - LookUp Table transform for YV12
// YV12LUT (C) 2003 Kurosu (kurosu@inforezo.org)

// Based on "Levels" function from AviSynth software

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include <stdio.h>
#include <string.h>
#include "Expressions.h"
#include "GenericTools.h"
#include "LUT.h"
#include "info.h"

double clip(double x)
{
	if (x < 0.5) return 0;
	if (x > 254.5) return 255;
	return x+0.5;
}

LUT::LUT(PClip _child, const char *yExpr, const char *uExpr, const char *vExpr,
		 int _X, int _Y, int _w, int _h, int y, int u, int v,
		 IScriptEnvironment* env) : GenericVideoFilter(_child), w(_w), h(_h),
		 offX(_X & 0xFFFFFFFE), offY(_Y & 0xFFFFFFFE), Y(y), U(u), V(v)
{
	if (!vi.IsYV12())
		env->ThrowError("YV12LUT: Needs YV12 data");
	
	if (w<0) w = vi.width;
	if (w>vi.width-offX) w = vi.width-offX;
	if (h<0) h = vi.height;
	if (h>vi.height-offY) h = vi.height-offY;
	w &= 0xFFFFFFFE;
	h &= 0xFFFFFFFE;

	char separator = ' ';

	__asm { emms }

	Expression *Yexpr = Expression::BuildExpression(yExpr,separator);
	Expression *Uexpr = Expression::BuildExpression(uExpr,separator);
	Expression *Vexpr = Expression::BuildExpression(vExpr,separator);

	YLUT = new unsigned char[256];
	ULUT = new unsigned char[256];
	VLUT = new unsigned char[256];
    	
	for ( int x = 0; x < 256; x++ )
	{
		YLUT[x] = (unsigned char)(clip(Yexpr->process((double)x)));
		ULUT[x] = (unsigned char)(clip(Uexpr->process((double)x)));
		VLUT[x] = (unsigned char)(clip(Vexpr->process((double)x)));
	}
	
	delete Yexpr;
	delete Uexpr;
	delete Vexpr;

}

LUT::~LUT()
{
	delete YLUT;
	delete ULUT;
	delete VLUT;
}

PVideoFrame __stdcall LUT::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame		src = child->GetFrame(n, env);

	env->MakeWritable(&src);
	const int			src_pitch_y = src->GetPitch(PLANAR_Y);
	const int			src_pitch_uv = src->GetPitch(PLANAR_U);
	unsigned char		*srcp_u = src->GetWritePtr(PLANAR_U);
	unsigned char		*srcp_v = src->GetWritePtr(PLANAR_V);
	unsigned char		*srcp_y = src->GetWritePtr(PLANAR_Y);

	const int			row_size = src->GetRowSize(PLANAR_Y);
	const int			height = src->GetHeight(PLANAR_Y);

	int i, j;

	switch (Y)
	{
	case OPERATION:
		for ( i = offY; i < offY + h; i++ )
		{
			for ( j = offX; j < offX + w; j++ )
				srcp_y[j] = YLUT[srcp_y[j]];
			srcp_y += src_pitch_y;
		}
		break;
	case COPY_SOURCE:
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(srcp_y + offX + offY * src_pitch_y, -Y, w, h, offX, offY, src_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		for ( i = offY / 2; i < (offY + h) / 2; i++ )
		{
			for ( j = offX / 2; j < (offX + w) / 2; j++ )
				srcp_u[j] = ULUT[srcp_u[j]];
			srcp_u += src_pitch_uv;
		}
		break;
	case COPY_SOURCE:
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(srcp_u + offX / 2 + offY * src_pitch_uv / 2, -U, w / 2, h / 2, offX / 2, offY / 2, src_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		for ( i = offY / 2; i < (offY + h) / 2; i++ )
		{
			for ( j = offX / 2; j < (offX + w) / 2; j++ )
				srcp_v[j] = VLUT[srcp_v[j]];
			srcp_v += src_pitch_uv;
		}
		break;
	case COPY_SOURCE:
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(srcp_v + offX / 2 + offY * src_pitch_uv / 2, -U, w / 2, h / 2, offX / 2, offY / 2, src_pitch_uv);
		break;
	}

	return src;
}