// MaskedMerge - Weighed merge of video planes depending on masks
// MaskedMerge (C) 2003 Kurosu (kurosu@inforezo.org)

// Similar functionnalities (but not directly linked to it) can be found
// in MergeLuma and MergeChroma from Avisynth (www.avisynth.org) by
// Sh0dan (Sh0dan@avisynth.org)

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MaskedMerge.h"
#include "GenericTools.h"
#include "MM_func.h"

MaskedMerge::MaskedMerge(PClip _child, PClip cl2, PClip cl3, int _X, int _Y, int _w, int _h,
						 int y, int u, int v, bool usemmx, IScriptEnvironment* env) : 
GenericVideoFilter(_child), clip2(cl2), clip3(cl3), w(_w), h(_h),
offX(_X & 0xFFFFFFFE), offY(_Y & 0xFFFFFFFE), Y(y), V(v), U(u)
{
	const VideoInfo& vi2 = clip2->GetVideoInfo();
	const VideoInfo& vi3 = clip3->GetVideoInfo();
	
	if (!Verify(Y)) env->ThrowError("MaskedMerge: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("MaskedMerge: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("MaskedMerge: Unvalid operation mode on V");

	if (!vi.IsYV12() || !vi2.IsYV12() || !vi3.IsYV12())
		env->ThrowError("MaskedMerge: needs YV12");

	if (vi2.width != vi.width || vi.height != vi2.height)
		env->ThrowError("MaskedMerge: Size of 2nd clip different from 1st");
	if (vi3.width != vi.width || vi.height != vi3.height)
		env->ThrowError("MaskedMerge: Size of 3rd clip different from 1st");
	
	if (offX<0 || offX>vi.width)
		env->ThrowError("MaskedMerge: unvalid horizontal start position");
	if (offY<0 || offY>vi.height)
		env->ThrowError("MaskedMerge: unvalid vertical start position");

	//SetCacheHints(0,0);

	if (w < 0) w = vi.width;
	if (w + offX > vi.width) w = vi.width - offX;
	if (h < 0) h = vi.height;
	if (h + offY > vi.height) h = vi.height - offY;
	w &= 0xFFFFFFFE;
	h &= 0xFFFFFFFE;

	if ((env->GetCPUFlags() & CPUF_MMX) && (!(w%8)) && usemmx ) MMergeY = MMerge_MMX;
	else MMergeY = MMerge_C;

	if ((env->GetCPUFlags() & CPUF_MMX) && (!(w%16)) && usemmx ) MMergeUV = MMerge_MMX;
	else MMergeUV = MMerge_C;
}

PVideoFrame __stdcall MaskedMerge::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			src1 = child->GetFrame(n, env);
	env->MakeWritable(&src1);
	unsigned char		*src1p_y = src1->GetWritePtr(PLANAR_Y);
	unsigned char		*src1p_u = src1->GetWritePtr(PLANAR_U);
	unsigned char		*src1p_v = src1->GetWritePtr(PLANAR_V);
	const int			src1_pitch_y = src1->GetPitch(PLANAR_Y);
	const int			src1_pitch_uv = src1->GetPitch(PLANAR_U);
	
	PVideoFrame			src2 = clip2->GetFrame(n, env);
	const unsigned char	*src2p_y = src2->GetReadPtr(PLANAR_Y);
	const unsigned char	*src2p_u = src2->GetReadPtr(PLANAR_U);
	const unsigned char	*src2p_v = src2->GetReadPtr(PLANAR_V);
	const int			src2_pitch_y = src2->GetPitch(PLANAR_Y);
	const int			src2_pitch_uv = src2->GetPitch(PLANAR_U);
	
	const int			row_size = src1->GetRowSize(PLANAR_Y);
	const int			height = src1->GetHeight(PLANAR_Y);
	
	PVideoFrame			mask = clip3->GetFrame(n, env);
	const unsigned char	*maskp_y = mask->GetReadPtr(PLANAR_Y);
	const unsigned char	*maskp_u = mask->GetReadPtr(PLANAR_U);
	const unsigned char	*maskp_v = mask->GetReadPtr(PLANAR_V);
	const int			mask_pitch_y = mask->GetPitch(PLANAR_Y);
	const int			mask_pitch_uv = mask->GetPitch(PLANAR_U);
	
	switch (Y)
	{
	case OPERATION:
		MMergeY(	src1p_y + offX + offY * src1_pitch_y,
					src2p_y + offX + offY * src2_pitch_y,
					maskp_y + offX + offY * mask_pitch_y,
					src1_pitch_y, src2_pitch_y, mask_pitch_y, w, h);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_y + offX + offY * src1_pitch_y, src1_pitch_y,
					src2p_y + offX + offY * src2_pitch_y, src2_pitch_y, w, h);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_y, -Y, w, h, offX, offY, src1_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		MMergeUV(	src1p_u + (offX>>1) + (offY>>1) * src1_pitch_uv,
					src2p_u + (offX>>1) + (offY>>1) * src2_pitch_uv,
					maskp_u + (offX>>1) + (offY>>1) * mask_pitch_uv,
					src1_pitch_uv, src2_pitch_uv, mask_pitch_uv, w>>1, h>>1);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_u + (offX>>1) + (offY>>1) * src1_pitch_uv, src1_pitch_uv,
					src2p_u + (offX>>1) + (offY>>1) * src2_pitch_uv, src2_pitch_uv, w>>1, h>>1);
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_u, -U, w / 2, h / 2, offX>>1, offY>>1, src1_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		MMergeUV(	src1p_v + (offX>>1) + (offY>>1) * src1_pitch_uv,
					src2p_v + (offX>>1) + (offY>>1)*src2_pitch_uv,
					maskp_v + (offX>>1) + (offY>>1)*mask_pitch_uv,
					src1_pitch_uv, src2_pitch_uv, mask_pitch_uv, w>>1, h>>1);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_v + (offX>>1) + (offY>>1) * src1_pitch_uv, src1_pitch_uv,
					src2p_v + (offX>>1) + (offY>>1) * src2_pitch_uv, src2_pitch_uv, w>>1, h>>1);
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_v, -V, w>>1, h>>1, offX>>1, offY>>1, src1_pitch_uv);
		break;
	}
	
	return src1;
}