// Motion Detection : create a mask from a basic motion detection
// Motion Detection (C) 2004 Manao (manao@melix.net)

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MotionMask.h"
#include "MD_Func.h"
#include "GenericTools.h"

MotionMask::MotionMask(PClip _child, unsigned int thY1, unsigned int thY2,
				   unsigned int thC1, unsigned int thC2, unsigned int thSD, 
				   int y, int u, int v, bool usemmx, IScriptEnvironment* env) :
GenericVideoFilter(_child), Yth1(thY1), Yth2(thY2), Cth1(thC1), Cth2(thC2), SDth(thSD), Y(y), U(u), V(v)
{
	if (!Verify(Y)) env->ThrowError("MotionMask: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("MotionMask: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("MotionMask: Unvalid operation mode on V");

	if (!vi.IsYV12())
		env->ThrowError("MotionMask: Needs YV12 data");
	
	if (thY1 > 255)
		env->ThrowError("MotionMask: first luma threshold not in the range [0..255]");
	if (thY2 > 255)
		env->ThrowError("MotionMask: second luma threshold not in the range [0..255]");
	if (thC1 > 255)
		env->ThrowError("MotionMask: first chroma threshold not in the range [0..255]");
	if (thC2 > 255)
		env->ThrowError("MotionMask: second chroma threshold not in the range [0..255]");
	if (thSD > 255)
		env->ThrowError("MotionMask: scene detection threshold not in the range [0..255]");
	
	if (thY1 > thY2)
		env->ThrowError("MotionMask: the first luma threshold should not be superior to the second one");
	if (thC1 > thC2)
		env->ThrowError("MotionMask: the first chroma threshold should not be superior to the second one");

	MD = MD_C;

	if ((env->GetCPUFlags() & CPUF_INTEGER_SSE) && ( !( vi.width % 8 )) && usemmx ) SD = SD_ISSE;
	else SD = SD_C;

	last_n = -2;
}

MotionMask::~MotionMask()
{
}

PVideoFrame __stdcall MotionMask::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			der = env->NewVideoFrame(vi);
	unsigned char		*derp_y = der->GetWritePtr(PLANAR_Y);
	unsigned char		*derp_u = der->GetWritePtr(PLANAR_U);
	unsigned char		*derp_v = der->GetWritePtr(PLANAR_V);
	const int			der_pitch_y = der->GetPitch(PLANAR_Y);
	const int			der_pitch_uv = der->GetPitch(PLANAR_U);

	PVideoFrame			src = child->GetFrame(n, env);

	const int			row_size = src->GetRowSize(PLANAR_Y);
	const int			height = src->GetHeight(PLANAR_Y);

	const unsigned char *srcp_y = src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcp_u = src->GetReadPtr(PLANAR_U);
	const unsigned char *srcp_v = src->GetReadPtr(PLANAR_V);
	const int			src_pitch_y = src->GetPitch(PLANAR_Y);
	const int			src_pitch_uv = src->GetPitch(PLANAR_U);

	PVideoFrame			srcp = child->GetFrame(n - 1, env);

	const unsigned char		*prevy = srcp->GetReadPtr(PLANAR_Y);
	const unsigned char		*prevu = srcp->GetReadPtr(PLANAR_U);
	const unsigned char		*prevv = srcp->GetReadPtr(PLANAR_V);

	if ( SD(srcp_y, prevy, src_pitch_y, row_size, height, SDth) )
	{
		switch (Y)
		{
			case OPERATION:
				memset(derp_y, 0, der_pitch_y*height);
				break;
			case COPY_SOURCE:
				env->BitBlt(derp_y,der_pitch_y,srcp_y,src_pitch_y,row_size,height);
				break;
			case NO_OPERATION:
				break;
			default:
				MemZoneSet(derp_y, -Y, row_size, height, 0, 0, der_pitch_y);
				break;
		}
		switch (U)
		{
			case OPERATION:
				memset(derp_u, 0, (der_pitch_uv*height) >> 1);
				break;
			case COPY_SOURCE:
				env->BitBlt(derp_u,	der_pitch_uv, srcp_u, src_pitch_uv,
							(row_size>>1), (height>>1));
			case NO_OPERATION:
				break;
			default:
				MemZoneSet(derp_u, -U, row_size / 2, height / 2, 0, 0, der_pitch_uv);
				break;
		}
			
		switch (V)
		{
			case OPERATION:
				memset(derp_v, 0, (der_pitch_uv*height) >> 1);
				break;
			case COPY_SOURCE:
				env->BitBlt(derp_v,	der_pitch_uv, srcp_v, src_pitch_uv,
							(row_size>>1),(height>>1));
				break;
			case NO_OPERATION:
				break;
			default:
				MemZoneSet(derp_v, -V, row_size / 2, height / 2, 0, 0, der_pitch_uv);
				break;
		}
	}
	else
	{

		switch (Y)

			{
			case OPERATION:
				MD(srcp_y, prevy, derp_y, src_pitch_y, src_pitch_y, der_pitch_y,	row_size,
					height,	Yth1, Yth2);
				break;
			case COPY_SOURCE:
				env->BitBlt(derp_y,der_pitch_y,srcp_y,src_pitch_y,row_size,height);
				break;
			case NO_OPERATION:
				break;
			default:
				MemZoneSet(derp_y, -Y, row_size, height, 0, 0, der_pitch_y);
				break;
			}
			
			switch (U)
			{
			case OPERATION:
				MD(srcp_u, prevu, derp_u, src_pitch_uv, src_pitch_uv, der_pitch_uv, row_size>>1,
					height>>1, Cth1, Cth2);
				break;
			case COPY_SOURCE:
				env->BitBlt(derp_u,	der_pitch_uv, srcp_u, src_pitch_uv,
							(row_size>>1), (height>>1));
			case NO_OPERATION:
				break;
			default:
				MemZoneSet(derp_u, -U, row_size / 2, height / 2, 0, 0, der_pitch_uv);
				break;
			}
			
			switch (V)
			{
			case OPERATION:
				MD(srcp_v, prevv, derp_v, src_pitch_uv, src_pitch_uv, der_pitch_uv, row_size>>1,
					height>>1, Cth1, Cth2);
				break;
			case COPY_SOURCE:
				env->BitBlt(derp_v,	der_pitch_uv, srcp_v, src_pitch_uv,
							(row_size>>1),(height>>1));
				break;
			case NO_OPERATION:
				break;
			default:
				MemZoneSet(derp_v, -V, row_size / 2, height / 2, 0, 0, der_pitch_uv);
				break;
		}
	}

	return der;
}
