// General Convolution : allow to make a convolution in YV12 colorspace
// General Convolution (C) 2004 Manao (manao@melix.net)

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include <string.h>
#include <stdio.h>
#include <math.h>
#include "YV12Convolution.h"
#include "GC_Func.h"
#include "GenericTools.h"

bool YV12Convolution::AnalyseString(const char *string, double **vectf, int **vecti, int *halflength, IScriptEnvironment* env, bool *isFloat)
{
	int separatorCount = 0;
	int length = strlen(string);
	char separator = ' ';
	bool answer = false;
	int i, count;

	for ( i = 0; i < length; i++ )
		separatorCount += (string[i] == separator) ? 1 : 0;

	const char *sp = string;
	char *array;

	(*isFloat) |= (!(NULL == strchr(string,'.')));

	(*halflength) = (separatorCount / 2);
	if (separatorCount % 2) env->ThrowError("YV12Convolution: Vector's length have to be odd");
	(*vectf) = new double[(*halflength)*2+1];
	(*vecti) = new int[(*halflength)*2+1];

	for ( count = 0; count < separatorCount; count++ )
	{
		int end = strcspn(sp,&separator);
		array = new char[end+1];
		for ( int i = 0; i < end; i++ )
			array[i] = sp[i];
		array[end] = 0;
		sp = sp + end + strspn(sp+end,&separator);

		(*vectf)[count] = atof(array);
		(*vecti)[count] = atof(array);
		answer |= ( (*vectf)[count] < 0 );
		delete[] array;
	}

	int end = strlen(sp);
	array = new char[end+1];
	for ( i = 0; i < end; i++ )
		array[i] = sp[i];
	array[end] = 0;

	(*vectf)[separatorCount] = atof(array);
	(*vecti)[separatorCount] = atof(array);
	answer |= ( (*vectf)[separatorCount] < 0 );
	delete[] array;
	
	return answer;
}

YV12Convolution::YV12Convolution(PClip _child, const char *horizontal, const char *vertical,
								int ttl, bool automatic, bool saturate, int y, int u, int v, bool usemmx,
								IScriptEnvironment* env) :
GenericVideoFilter(_child), Y(y), U(u), V(v)
{
	if (!Verify(Y)) env->ThrowError("YV12Convolution: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("YV12Convolution: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("YV12Convolution: Unvalid operation mode on V");
 
	if (!vi.IsYV12())
		env->ThrowError("YV12Convolution: Needs YV12 data");

	total = ttl;
	totalf = ttl;
	bool isNegative = saturate;
	bool isFloat = false;

	isNegative |= AnalyseString(horizontal,&horvectf,&horvect,&horres,env,&isFloat);
	isNegative |= AnalyseString(vertical,&vervectf,&vervect,&verres,env,&isFloat);

	if (isFloat)
	{
		double other_total = 0;

		if (automatic) {
			other_total = 0;
			for (int x = 0; x < horres * 2 + 1; x++ )
				for (int y = 0; y < verres * 2 + 1; y++ )
					other_total += horvectf[x] * vervectf[y];
		}

		if ( isNegative ) GC = GCf_C_ABS;
		else if (( !automatic ) && ( totalf < other_total )) GC = GCf_C_SAT;
		else GC = GCf_C;
		
		if ( automatic ) totalf = other_total;
		if ( fabs(totalf) < 0.001 ) totalf = 1.0;
	}
	else {
		int other_total = 0;

		if (automatic) {
			other_total = 0;
			for (int x = 0; x < horres * 2 + 1; x++ )
				for (int y = 0; y < verres * 2 + 1; y++ )
					other_total += horvect[x] * vervect[y];
		}

		if ( isNegative ) GC = GC_C_ABS;
		else if (( !automatic ) && ( total < other_total )) GC = GC_C_SAT;
		else GC = GC_C;

		if ( automatic ) total = other_total;

		if ( total == 0 ) total = 1;
	}


}

YV12Convolution::~YV12Convolution()
{
	delete[] horvect;
	delete[] vervect;
	delete[] horvectf;
	delete[] vervectf;
}

PVideoFrame __stdcall YV12Convolution::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			der = env->NewVideoFrame(vi);
	env->MakeWritable(&der);

	unsigned char		*derp_y = der->GetWritePtr(PLANAR_Y);
	unsigned char		*derp_u = der->GetWritePtr(PLANAR_U);
	unsigned char		*derp_v = der->GetWritePtr(PLANAR_V);
	const int			der_pitch_y = der->GetPitch(PLANAR_Y);
	const int			der_pitch_uv = der->GetPitch(PLANAR_U);

	PVideoFrame			src = child->GetFrame(n, env);	

	const int			row_size = src->GetRowSize(PLANAR_Y);
	const int			height = src->GetHeight(PLANAR_Y);

	const unsigned char *srcp_y = src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcp_u = src->GetReadPtr(PLANAR_U);
	const unsigned char *srcp_v = src->GetReadPtr(PLANAR_V);
	const int			src_pitch_y = src->GetPitch(PLANAR_Y);
	const int			src_pitch_uv = src->GetPitch(PLANAR_U);

	switch (Y)

	{
		case OPERATION:
			GC(srcp_y, derp_y, src_pitch_y, der_pitch_y, row_size, height, horvect, vervect, horvectf, vervectf,
				horres, verres, total, totalf);
			break;
		case COPY_SOURCE:
			env->BitBlt(derp_y, der_pitch_y, srcp_y, src_pitch_y, row_size, height);
			break;
		case NO_OPERATION:
			break;
		default:
			MemZoneSet(derp_y, -Y, row_size, height, 0, 0, der_pitch_y);
			break;
		}
		
		switch (U)
		{
		case OPERATION:
			GC(srcp_u, derp_u, src_pitch_uv, der_pitch_uv, row_size >> 1, height >> 1, horvect, vervect, horvectf, vervectf,
				horres, verres, total, totalf);
			break;
		case COPY_SOURCE:
			env->BitBlt(derp_u, der_pitch_uv, srcp_u, src_pitch_uv, row_size >> 1, height >> 1);
			break;
		case NO_OPERATION:
			break;
		default:
			MemZoneSet(derp_u, -U, row_size / 2, height / 2, 0, 0, der_pitch_uv);
			break;
		}
		
		switch (V)
		{
		case OPERATION:
			GC(srcp_v, derp_v, src_pitch_uv, der_pitch_uv, row_size >> 1, height >> 1, horvect, vervect, horvectf, vervectf,
				horres, verres, total, totalf);
			break;
		case COPY_SOURCE:
			env->BitBlt(derp_v, der_pitch_uv, srcp_v, src_pitch_uv, row_size >> 1, height >> 1);
			break;
		case NO_OPERATION:
			break;
		default:
			MemZoneSet(derp_v, -V, row_size / 2, height / 2, 0, 0, der_pitch_uv);
			break;
	}

	return der;
}
