// YV12Subtract - expansion of AviSynth's Subtract
// YV12Subtract (C) 2003 Kurosu (kurosu@inforezo.org)
// Subtract (C) AviSynth's team

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "GenericTools.h"
#include "YV12Substract.h"
#include "YV12S_Func.h"
#include <math.h>

YV12Substract::YV12Substract(PClip _child, PClip _child2, double _tol, int _X, int _Y,
							 int _w, int _h, int y, int u, int v, bool wideRange, bool useMMX, IScriptEnvironment* env) :
GenericVideoFilter(_child), child2(_child2), Y(y), V(v), U(u), w(_w), h(_h),
offX(_X & 0xFFFFFFFE), offY(_Y & 0xFFFFFFFE), tolerance(_tol)
{
	const VideoInfo& vi1 = _child->GetVideoInfo();
	const VideoInfo& vi2 = _child2->GetVideoInfo();
	if ( (vi1.width != vi2.width) || (vi1.height != vi2.height) )
		env->ThrowError("YV1Substract: Image dimensions don't match");
	/*if (vi1.pixel_type != vi2.pixel_type)
		env->ThrowError("YV1Substract: Image formats don't match");*/
	if (!vi1.IsYV12() || !vi2.IsYV12())
		env->ThrowError("YV12Substract: requires YV12 colorspace");
	
	if (offX<0 || offX>vi.width)
		env->ThrowError("YV12Substract: horizontal offset %i bigger than source width %i",
						offX, vi.width);
	if (offY<0 || offY>vi.height)
		env->ThrowError("YV12Substract: vertical offset %i bigger than source height %i",
						offY, vi.height);

	if (w<0) w = vi.width;
	if (w>vi.width-offX) w = vi.width-offX;
	if (h<0) h = vi.height;
	if (h>vi.height-offY) h = vi.height-offY;

	if ( wideRange )
	{
		double alpha;
		if ( _tol <= 0 ) alpha = 0.5;
		else alpha = 1.0 / _tol;
		LUT = new unsigned char[511];
		for ( int i = 0; i < 511; i++ )
		{
			double x = i - 255;
			if ( i < 255 ) LUT[i] = 128.0 - 128.0 * pow(-x / 255.0, alpha);
			else LUT[i] = 128.0 + 127.0 * pow(x / 255.0, alpha);
		}
		if ((env->GetCPUFlags() & CPUF_MMX) && useMMX)
		{
			if (!(vi1.width%16) && !(vi1.height%4))
			{SubstractY = SubstractW_MMX; SubstractUV = SubstractW_MMX; }
			else if (!(vi1.width%8) && !(vi1.height%4))
			{SubstractY = SubstractW_MMX; SubstractUV = SubstractW_C2; }
			else if (!(vi1.width%8) && !(vi1.height%2))
			{SubstractY = SubstractW_MMX; SubstractUV = SubstractW_C1; }
			else if (!(vi1.width%4) && !(vi1.height%4))
			{SubstractY = SubstractW_C2; SubstractUV = SubstractW_C2; }
			else {SubstractY = SubstractW_C2; SubstractUV = SubstractW_C1; }
		}
		else if (!(vi1.width%4) && !(vi1.height%4))
		{SubstractY = SubstractW_C2; SubstractUV = SubstractW_C2; }
		else {SubstractY = SubstractW_C2; SubstractUV = SubstractW_C1; }
	}
	else if (_tol<0)
	{
		if ((env->GetCPUFlags() & CPUF_MMX) && useMMX)
		{
			if (!(vi1.width%16) && !(vi1.height%4))
			{SubstractY = Substract_MMX; SubstractUV = Substract_MMX; }
			else if (!(vi1.width%8) && !(vi1.height%4))
			{SubstractY = Substract_MMX; SubstractUV = Substract_C2; }
			else if (!(vi1.width%8) && !(vi1.height%2))
			{SubstractY = Substract_MMX; SubstractUV = Substract_C1; }
			else if (!(vi1.width%4) && !(vi1.height%4))
			{SubstractY = Substract_C2; SubstractUV = Substract_C2; }
			else {SubstractY = Substract_C2; SubstractUV = Substract_C1; }
		}
		else if (!(vi1.width%4) && !(vi1.height%4))
		{SubstractY = Substract_C2; SubstractUV = Substract_C2; }
		else {SubstractY = Substract_C2; SubstractUV = Substract_C1; }
	}
	else
	{
		if (_tol>255)
			env->ThrowError("YV12Substract: tolerance %i isn't in [0..255]",_tol);
		if ((env->GetCPUFlags() & CPUF_MMX) && useMMX)
		{
			if (!(vi1.width%16) && !(vi1.height%4))
			{SubstractY = Diff_MMX; SubstractUV = Diff_MMX; }
			else if (!(vi1.width%8) && !(vi1.height%2))
			{SubstractY = Diff_MMX; SubstractUV = Diff_C; }
			else {SubstractY = Diff_C; SubstractUV = Diff_C; }
		}
		else {SubstractY = Diff_C; SubstractUV = Diff_C; }
	}
	if (!Verify(Y)) env->ThrowError("YV12Substract: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("YV12Substract: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("YV12Substract: Unvalid operation mode on V");
}

PVideoFrame __stdcall YV12Substract::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			src1 = child->GetFrame(n, env);
	env->MakeWritable(&src1);
	unsigned char		*src1p_y = src1->GetWritePtr(PLANAR_Y);
	unsigned char		*src1p_u = src1->GetWritePtr(PLANAR_U);
	unsigned char		*src1p_v = src1->GetWritePtr(PLANAR_V);
	const int			src1_pitch_y = src1->GetPitch(PLANAR_Y);
	const int			src1_pitch_uv = src1->GetPitch(PLANAR_U);
	
	PVideoFrame			src2 = child2->GetFrame(n, env);
	const unsigned char	*src2p_y = src2->GetReadPtr(PLANAR_Y);
	const unsigned char	*src2p_u = src2->GetReadPtr(PLANAR_U);
	const unsigned char	*src2p_v = src2->GetReadPtr(PLANAR_V);
	const int			src2_pitch_y = src2->GetPitch(PLANAR_Y);
	const int			src2_pitch_uv = src2->GetPitch(PLANAR_U);
	
	const int			row_size = src1->GetRowSize(PLANAR_Y);
	const int			height = src1->GetHeight(PLANAR_Y);
	
	switch (Y)
	{
	case OPERATION:
		SubstractY(src1p_y + offX + offY*src1_pitch_y, src2p_y + offX + offY*src2_pitch_y,
					src1_pitch_y, src2_pitch_y, w, h, tolerance, LUT);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_y + offX + offY * src1_pitch_y, src1_pitch_y,
					src2p_y + offX + offY * src2_pitch_y, src2_pitch_y, w, h);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_y, -Y, w, h, offX, offY, src1_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		SubstractUV(src1p_u + (offX>>1) + (offY>>1) * src1_pitch_uv,
					src2p_u + (offX>>1) + (offY>>1) * src2_pitch_uv,
					src1_pitch_uv, src2_pitch_uv, w>>1, h>>1, tolerance, LUT);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_u + (offX>>1) + (offY>>1) * src1_pitch_uv, src1_pitch_uv,
					src2p_u + (offX>>1) + (offY>>1) * src2_pitch_uv, src2_pitch_uv, w>>1, h>>1);
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_u, -U, w>>1, h>>1, offX>>1, offY>>1, src1_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		SubstractUV(src1p_v + (offX>>1) + (offY>>1) * src1_pitch_uv,
					src2p_v + (offX>>1) + (offY>>1) * src2_pitch_uv,
					src1_pitch_uv, src2_pitch_uv, w>>1, h>>1, tolerance, LUT);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_v + (offX>>1) + (offY>>1) * src1_pitch_uv, src1_pitch_uv,
					src2p_v + (offX>>1) + (offY>>1) * src2_pitch_uv, src2_pitch_uv, w>>1, h>>1);
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_v, -V, w>>1, h>>1, offX>>1, offY>>1, src1_pitch_uv);
		break;
	}
	
	return src1;
}
